"""General utilities for the diagnostic tools."""

import numpy as np

# Some constants
timing_resolution_fine = 1.5625e-9

DIV = np.uint32(0x2000)

# Tristan 10M specs
tristan_modules = {"10M": (2, 5), "2M": (1, 2), "1M": (1, 1)}  # (H, V) -.> (fast, slow)
mod_size = (515, 2069)  # slow, fast
gap_size = (117, 45)  # slow, fast
image_size = (3043, 4183)  # slow, fast


def define_modules(num_modules: str = "10M") -> dict:
    """Define the start and end pixel of each module in the Tristan detector.

    Args:
        num_modules (str, optional): Specify how many modules make up the Tristan detector currently in use. Defaults to "10M".

    Returns:
        dict[tuple, tuple]: Start and end pixel value of each module - which are defined by a (x,y) tuple.
            For example a Tristan 1M will return {(0,0): {"10M": (2, 5), "2M": (1, 2), "1M": (1,1)}}
    """
    modules = tristan_modules[num_modules]
    mod = {}
    for _y in range(modules[0]):
        for _x in range(modules[1]):
            int_x = [
                _x * (mod_size[0] + gap_size[0]),
                _x * (mod_size[0] + gap_size[0]) + mod_size[0],
            ]
            int_y = [
                _y * (mod_size[1] + gap_size[1]),
                _y * (mod_size[1] + gap_size[1]) + mod_size[1],
            ]
            mod[(_x, _y)] = (int_x, int_y)
    return mod



"""Tools for binning events to images."""
from __future__ import annotations

from operator import mul
from pathlib import Path
from typing import Sequence

import numpy as np
import pandas as pd
import zarr
from dask import dataframe as dd
from dask import distributed
from numpy.typing import ArrayLike

from .data import (
    cue_id_key,
    cue_time_key,
    event_location_key,
    event_time_key,
    pixel_index,
    shutter_close,
    shutter_open,
    valid_events,
)


def find_start_end(data: dd.DataFrame) -> (int, int):
    """
    Find the shutter open and shutter close timestamps.

    Args:
        data:  LATRD data.  Must contain one 'cue_id' entry and one
               'cue_timestamp_zero' entry.  The two arrays are assumed to have the
               same length.

    Returns:
        The shutter open and shutter close timestamps, in clock cycles.
    """
    indices = (data[cue_id_key] == shutter_open) | (data[cue_id_key] == shutter_close)
    times = data[cue_time_key][indices]
    start, end = np.unique(times.compute())

    return start, end


def align_bins(start: int, align: int, end: int, n_bins: int) -> (int, int):
    """
    Divide an interval into a specified number of bins, aligning with a given value.

    Take three integers, ``start`` ≤ ``align`` ≤ ``end``, and find a way to span the
    largest possible interval between ``start`` and ``end`` with a specified number of
    bins, while ensuring that one of the bin edges is aligned with a specified value.

    Args:
        start:   The start of the interval.
        align:   The value to which a bin edge should be aligned.
        end:     The end of the interval.
        n_bins:  The number of bins.

    Returns:
        The first bin edge and the bin width, from which all the bin edges can be
        derived.
    """
    if align <= (start + end) / 2:
        # Find the bin width by pinning a bin edge to the align time and the last bin
        # edge to the end time, then maximising the number of bins we can fit between
        # the start time and the trigger time.
        # At least half the images must happen after the trigger time.
        n_bins_after = np.arange(n_bins // 2 or 1, n_bins + 1)
        bin_width = ((end - align) / n_bins_after).astype(int)
        new_start = end - n_bins * bin_width
        best = np.argmax(new_start >= start)
        start = new_start[best]
        bin_width = bin_width[best]
    else:
        # Find the bin width by pinning a bin edge to the align time and the first
        # bin edge to the start time, then maximising the number of bins we can fit
        # between the trigger time and the end time.
        # At least half the images must happen before the trigger time.
        n_bins_before = np.arange(n_bins // 2 or 1, n_bins + 1)
        bin_width = ((align - start) / n_bins_before).astype(int)
        new_end = start + n_bins * bin_width
        best = np.argmax(new_end <= end)
        bin_width = bin_width[best]

    return start, bin_width


def create_cache(
    output_file: Path | str, num_images: int, image_size: tuple[int, int]
) -> zarr.Array:
    """
    Make a Zarr array of zeros, suitable for using as an image binning cache.

    The array will have shape (num_images, *image_size) and will be chunked by image,
    i.e. the chunk shape will be (1, *image_size).

    Args:
        output_file:  Output file name.  Any file extension will be replaced with .zarr.
        num_images:   The number of images in the array.
        image_size:   The size of an image in the array.

    Returns:

    """
    shape = num_images, *image_size
    chunks = 1, *image_size
    output = Path(output_file).with_suffix(".zarr")
    return zarr.zeros(
        shape, chunks=chunks, dtype=np.int32, overwrite=True, store=output, path="data"
    )


def find_time_bins(data: pd.DataFrame, bins: Sequence[int]):
    """
    Convert the event timestamps in LATRD data to time bin indices.

    For each event, determine the index of the bin into which the event will fall.

    Args:
        data:  LATRD events data.  Must have an ``event_time_offset`` column.
        bins:  The time bin edges of the images (in clock cycles, to match the event
               timestamps).

    Returns:
        A DataFrame which matches the input data except that the
        ``event_time_offset`` column is replaced with a column of ``time_bin`` indices.
    """
    num_images = len(bins) - 1

    # Find the index of the image to which each event belongs.
    if num_images > 1:
        data[event_time_key] = np.digitize(data[event_time_key], bins) - 1
    elif num_images:
        data[event_time_key] = 0

    return data.rename(columns={event_time_key: "time_bin"})


def make_images(data: pd.DataFrame, image_size: tuple[int, int], cache: ArrayLike):
    """
    Bin LATRD events data into images of event counts.

    Given a collection of events data, a known image shape and an array of the
    desired time bin edges, make an image for each time bin, representing the number
    of events recorded at each pixel.  Add the binned images to an array representing
    the full image stack.

    Args:
        data:        LATRD data.  Must have an ``event_id`` column and an
                    ``image_index`` column.
        image_size:  The (y, x), i.e. (slow, fast) dimensions (number of pixels) of
                     the image.
        cache:       Array representing the image stack, to which the binned events
                     should be added.  This might be a Zarr array, in which case it
                     functions as an on-disk cache of the binned images.
    """
    # Construct a stack of images using dask.array.bincount and add them to the cache.
    for image_index in data["time_bin"].unique():
        image_index = int(image_index)  # Convert to a serialisable type.
        locations = data[event_location_key][data["time_bin"] == image_index]
        pixel_counts = np.bincount(locations, minlength=mul(*image_size))
        pixel_counts = pixel_counts.astype(np.uint32).reshape(image_size)
        with distributed.Lock(image_index):
            # Beware!  When the cache is a Zarr array, using inplace addition (+=)
            # here means the locking fails to prevent the race condition of multiple
            # make_images calls in separate threads accessing the same chunk of the
            # cache simultaneously.  Presumably this is due to Zarr releasing the GIL
            # before it has finished flushing the result of the inplace addition to
            # disk.
            cache[image_index] = cache[image_index] + pixel_counts

    return pd.DataFrame(columns=data.columns)


def events_to_images(
    data: dd.DataFrame,
    bins: Sequence[int],
    image_size: tuple[int, int],
    cache: ArrayLike,
) -> dd.DataFrame:
    """
    Construct a stack of images from events data.

    From a sequence of LATRD events data, bin the events to images and store the
    binned images in a cache array.  The cache may be backed with on-disk storage,
    as in the case of a Zarr array, or may be a simple in-memory object, like a NumPy
    array.

    Args:
        data:        LATRD events data.  Must have an ``event_time_offset`` column
                     and an ``event_id`` column.
        bins:        The time bin edges of the images (in clock cycles, to match the
                     event timestamps).
        image_size:  The size of each image.
        cache:       An array representing the eventual image stack, having shape
                     ``(len(bins) - 1, *image_size)``, to which the pixel counts from
                     this binning operation will be added.

    Returns:
        A Dask collection representing the lazy image binning computation.
    """
    # Consider only those events that occur between the start and end times.
    data = valid_events(data, bins[0], bins[-1])
    # Convert the event IDs to a form that is suitable for a NumPy bincount.
    data[event_location_key] = pixel_index(data[event_location_key], image_size)

    columns = event_location_key, "time_bin"
    dtypes = data.dtypes
    dtypes["time_bin"] = dtypes.pop(event_time_key)
    meta = pd.DataFrame(columns=columns).astype(dtype=dtypes)
    data = data.map_partitions(find_time_bins, bins=bins, meta=meta)

    # Bin to images, partition by partition.
    data = dd.map_partitions(
        make_images, data, image_size, cache, meta=meta, enforce_metadata=False
    )

    return data



"""
Check that all files from all detector modules contain valid data.
"""
from __future__ import annotations

epilog_message = """
This program runs through all the files written for a Tristan collection and checks that they contain events, \
as well as that they are assigned to the correct module.\n
The results are written to a filename_MODULECHECK.log.
"""

import argparse
import glob
import logging
import time
from pathlib import Path

import h5py

from ..command_line import version_parser
from ..data import event_location_key
from . import DIV, define_modules
from . import diagnostics_log as log

# Define a logger
logger = logging.getLogger("TristanDiagnostics.ModuleCheck")

# Define parser
usage = "%(prog)s /path/to/data/dir filename_root [options]"
parser = argparse.ArgumentParser(
    usage=usage,
    formatter_class=argparse.RawTextHelpFormatter,
    description=__doc__,
    epilog=epilog_message,
    parents=[version_parser],
)
parser.add_argument("visitpath", type=str, help="Visit directory.")
parser.add_argument("filename", type=str, help="Root filename.")
parser.add_argument(
    "-l",
    "--list",
    action="store_true",
    help="Print out the list of files corresponding to each module.",
)
parser.add_argument(
    "-o",
    "--output",
    type=str,
    help="Output directory to save results/log file. If not passed, the script will default to current working directory.",
)
parser.add_argument(
    "-m",
    "--num-modules",
    choices=["1M", "2M", "10M"],
    default="10M",
    type=str,
    help="Number of detector modules.",
)


def main(args):
    filepath = Path(args.visitpath).expanduser().resolve()
    base = args.filename + f"_{6*'[0-9]'}.h5"

    if args.output:
        savedir = Path(args.output).expanduser().resolve()
        savedir.mkdir(exist_ok=True)
    else:
        savedir = Path.cwd()

    logfile = savedir / (filepath.stem + "_MODULECHECK.log")
    log.config(logfile.as_posix())

    logger.info(f"Quick data check for Tristan {args.num_modules} modules.")
    logger.info(f"Collection directory: {filepath}")
    logger.info(f"Filename root: {args.filename}")

    filename_template = filepath / base
    file_list = [
        Path(f).expanduser().resolve()
        for f in sorted(glob.glob(filename_template.as_posix()))
    ]
    logger.info(f"Found {len(file_list)} files in directory.")

    MOD = define_modules(args.num_modules)
    logger.info("Assigning each data file to correct module.\n")
    split = {k: [] for k in MOD.keys()}
    broken = []
    for filename in file_list:
        with h5py.File(filename) as fh:
            try:
                # Note: checking item of index 1 because for broken files there will
                # just be one item in "event_id" set to 0.
                x, y = divmod(fh[event_location_key][1], DIV)
                for k, v in MOD.items():
                    if v[1][0] <= x <= v[1][1]:
                        if v[0][0] <= y <= v[0][1]:
                            split[k].append(filename.name)
            except IndexError:
                broken.append(filename.name)

    num = 0
    for k, v in split.items():
        logger.info(f"--- Module {num} ---")
        logger.info(f"Position on detector: {k}")
        logger.info(f"Number of files found for this module: {len(v)}")
        num += 1
        if args.list:
            for f in v:
                logger.info(f"{f}")
        logger.info("\n")

    if len(broken) == 0.0:
        logger.info("No broken files found.")
    else:
        logger.warning(f"{len(broken)} broken files found.")
        logger.info("The following files all have missing data:")
        for i in broken:
            logger.info(f"{i}")


def cli():
    tic = time.time()
    args = parser.parse_args()
    main(args)
    toc = time.time()
    logger.debug(f"Total time taken: {toc - tic:.4f} s.")



"""Tools for extracting data on cues and events from Tristan data files."""

from __future__ import annotations

import re
from contextlib import ExitStack, contextmanager
from pathlib import Path
from typing import Iterable

import dask
import h5py
import numpy as np
from dask import array as da
from dask import dataframe as dd
from numpy.typing import ArrayLike
from pint import Quantity

from . import clock_frequency, ureg

# Regex for the names of data sets, in the time slice metadata file, representing the
# distribution of time slices across raw data files for each module.
ts_key_regex = re.compile(r"ts_qty_module\d{2}")

# Translations of the basic cue_id messages.
padding = np.uint16(0)
extended_timestamp = np.uint16(0x800)
shutter_open = np.uint16(0x840)
shutter_close = np.uint16(0x880)
fem_falling = np.uint16(0x8C1)
fem_rising = np.uint16(0x8E1)
ttl_falling = np.uint16(0x8C9)
ttl_rising = np.uint16(0x8E9)
lvds_falling = np.uint16(0x8CA)
lvds_rising = np.uint16(0x8EA)
tzero_falling = np.uint16(0x8CB)
tzero_rising = np.uint16(0x8EB)
sync_falling = np.uint16(0x8CC)
sync_rising = np.uint16(0x8EC)
reserved = np.uint16(0xF00)
cues = {
    padding: "Padding",
    extended_timestamp: "Extended time stamp, global synchronisation",
    shutter_open: "Shutter open time stamp, global",
    shutter_close: "Shutter close time stamp, global",
    fem_falling: "FEM trigger, falling edge",
    fem_rising: "FEM trigger, rising edge",
    ttl_falling: "Clock trigger TTL input, falling edge",
    ttl_rising: "Clock trigger TTL input, rising edge",
    lvds_falling: "Clock trigger LVDS input, falling edge",
    lvds_rising: "Clock trigger LVDS input, rising edge",
    tzero_falling: "Clock trigger TZERO input, falling edge",
    tzero_rising: "Clock trigger TZERO input, rising edge",
    sync_falling: "Clock trigger SYNC input, falling edge",
    sync_rising: "Clock trigger SYNC input, rising edge",
    np.uint16(0xBC6): "Error: messages out of sync",
    np.uint16(0xBCA): "Error: messages out of sync",
    reserved: "Reserved",
    **{
        basic + n: f"{name} time stamp, sensor module {n}"
        for basic, name in (
            (extended_timestamp, "Extended"),
            (shutter_open, "Shutter open"),
            (shutter_close, "Shutter close"),
        )
        for n in np.arange(1, 64, dtype=np.uint16)
    },
}

# Keys of cues and events data in the HDF5 file structure.
cue_id_key = "cue_id"
cue_time_key = "cue_timestamp_zero"
event_location_key = "event_id"
event_time_key = "event_time_offset"
event_energy_key = "event_energy"

cue_keys = cue_id_key, cue_time_key
event_keys = event_location_key, event_time_key, event_energy_key

nx_size_key = "entry/instrument/detector/module/data_size"


@contextmanager
def latrd_data(
    raw_file_paths: Iterable[str | Path],
    keys: Iterable[str] = cue_keys + event_keys,
) -> dd.DataFrame | dict[str, da.Array]:
    """
    A context manager to read LATRD data sets from multiple files.

    The yielded DataFrame has a column for each of the specified LATRD data keys.
    Each key must be a valid LATRD data key and the chosen data sets must all have the
    same length.  The data will be rechunked into partitions approximately the size of
    the default Dask array chunk size, but with chunk boundaries aligned with HDF5
    file boundaries.

    Args:
        raw_file_paths:  The paths of the raw LATRD data files.
        keys:  The set of LATRD data keys to be read.

    Yields:
        The data from all the files.
    """
    with ExitStack() as stack:
        files = [stack.enter_context(h5py.File(p, swmr=True)) for p in raw_file_paths]

        # Determine an appropriate block size for a Dask DataFrame of these data,
        # remembering to leave room for a 64-bit index.
        row_size = sum(files[0][k].dtype.itemsize for k in keys)
        block_size = ureg.Quantity(dask.config.get("array.chunk-size"))
        block_length = int(block_size.m_as("B") / row_size)

        # Construct a single Dask DataFrame from the specified keys.
        data = {
            k: da.concatenate([da.from_array(f[k], chunks=block_length) for f in files])
            for k in keys
        }
        data = dd.concat(
            [dd.from_dask_array(v, columns=k) for k, v in data.items()], axis=1
        )

        yield data


def first_cue_time(
    data: dd.DataFrame, message: int, after: int | None = None
) -> dd.DataFrame | None:
    """
    Find the timestamp of the first instance of a cue message in a Tristan data set.

    Args:
        data:     LATRD data.  Must contain one 'cue_id' column and one
                  'cue_timestamp_zero' column.  The two arrays are assumed to have the
                  same length.
        message:  The message code, as defined in the Tristan standard.
        after:    Ignore instances of the specified message before this timestamp.

    Returns:
        The timestamp, measured in clock cycles from the global synchronisation signal.
        If the message doesn't exist in the data set, this returns None.
    """
    message_incidences = data[cue_id_key] == message
    if after:
        message_incidences &= data[cue_time_key] >= after
    first_index = message_incidences.idxmax().compute()
    if first_index or data[cue_id_key].loc[0].compute().values == message:
        return data[cue_time_key].loc[first_index]


def cue_times(
    data: dd.DataFrame,
    message: int,
    after: int | None = None,
    before: int | None = None,
) -> da.Array:
    """
    Find the timestamps of all instances of a cue message in a Tristan data set.

    The found timestamps are de-duplicated.

    Args:
        data:     A DataFrame of LATRD data.  Must contain one column for cue id
                  messages and one for cue timestamps.
        message:  The message code, as defined in the Tristan standard.
        after:    Ignore instances of the specified message before this timestamp.

    Returns:
        The timestamps, measured in clock cycles from the global synchronisation
        signal, de-duplicated.
    """
    index = data[cue_id_key] == message
    if after:
        index &= data[cue_time_key] >= after
    if before:
        index &= data[cue_time_key] <= before
    return da.unique(data[cue_time_key][index].values)


def seconds(timestamp: int, reference: int = 0) -> Quantity:
    """
    Convert a Tristan timestamp to seconds, measured from a given reference timestamp.

    The time between the provided timestamp and a reference timestamp, both provided
    as a number of clock cycles from the same time origin, is converted to units of
    seconds.  By default, the reference timestamp is zero clock cycles, the beginning
    of the detector epoch.

    Args:
        timestamp:  A timestamp in number of clock cycles, to be converted to seconds.
        reference:  A reference time stamp in clock cycles.

    Returns:
        The difference between the two timestamps in seconds.
    """
    return ((timestamp - reference) / clock_frequency).to_base_units().to_compact()


def valid_events(data: dd.DataFrame, start: int, end: int) -> dd.DataFrame:
    """
    Return those events that have a timestamp in the specified range.

    Args:
        data:   LATRD data, containing an 'event_time_offset' column and optional
                'event_id' and 'event_energy' columns.
        start:  The start time of the accepted range, in clock cycles.
        end:    The end time of the accepted range, in clock cycles.

    Returns:
        The valid events.
    """
    valid = (start <= data[event_time_key]) & (data[event_time_key] < end)

    return data[valid]


def pixel_index(location: ArrayLike, image_size: tuple[int, int]) -> ArrayLike:
    """
    Extract pixel coordinate information from an event location (event_id) message.

    Translate a Tristan event location message to the index of the corresponding
    pixel in the flattened image array (i.e. numbered from zero, in row-major order).

    The pixel coordinates of an event on a Tristan detector are encoded in a 32-bit
    integer location message (the event_id) with 26 bits of useful information.
    Extract the y coordinate (the 13 least significant bits) and the x coordinate
    (the 13 next least significant bits).  Find the corresponding pixel index in the
    flattened image array by multiplying the y value by the size of the array in x,
    and adding the x value.

    This function calls the Python built-in divmod and so can be broadcast over
    array-like data structures.

    Args:
        location:    Event location message (an integer).
        image_size:  Shape of the image array in (y, x), i.e. (slow, fast).

    Returns:
        Index in the flattened image array of the pixel where the event occurred.
    """
    x, y = divmod(location, np.uint32(0x2000))
    # The following is equivalent to, but a little simpler than,
    # return da.ravel_multi_index((y, x), image_size)
    return x + y * image_size[1]



"""
Utilities for processing data from the Large Area Time-Resolved Detector

This module provides tools to interpret NeXus-like data in HDF5 format from the
experimental Timepix-based event-mode detector, codenamed Tristan, at Diamond Light
Source.
"""

from __future__ import annotations

__author__ = "Diamond Light Source — Data Analysis Group"
__email__ = "dataanalysis@diamond.ac.uk"
__version__ = "0.2.2"
__version_tuple__ = tuple(int(x) for x in __version__.split("."))

import dask
import pint
from dask.distributed import progress, wait

ureg = pint.UnitRegistry()

clock_frequency = ureg.Quantity(6.4e8, "Hz").to_compact()


def compute_with_progress(collection):
    """
    Compute a Dask collection, showing the progress of the top layer of the task graph.

    Args:
        collection:  A single Dask collection.
    """
    (collection,) = dask.persist(collection)

    # View progress only of the top layer of the task graph, which consists
    # of the rate limiting make_images tasks, to avoid giving a false sense
    # of rapid progress from the quick execution of the large number of
    # other, cheaper tasks.
    *_, top_layer = collection.dask.layers.values()
    futures = list(top_layer.values())
    print(progress(futures) or "")

    wait(collection)



"""Aggregate the events from a LATRD Tristan data collection into one or more images."""

from __future__ import annotations

import argparse
import sys
from contextlib import ExitStack
from operator import mul
from pathlib import Path

import dask
import h5py
import numpy as np
import pandas as pd
import pint
import zarr
from dask import array as da
from dask import dataframe as dd
from dask.diagnostics import ProgressBar
from dask.distributed import Client
from hdf5plugin import Bitshuffle
from nexgen.nxs_copy import CopyTristanNexus

from .. import clock_frequency, compute_with_progress
from ..binning import (
    align_bins,
    create_cache,
    events_to_images,
    find_start_end,
    find_time_bins,
    make_images,
)
from ..data import (
    cue_keys,
    cue_times,
    cues,
    event_location_key,
    event_time_key,
    first_cue_time,
    latrd_data,
    pixel_index,
    seconds,
    valid_events,
)
from . import (
    check_multiple_output_files,
    check_output_file,
    data_files,
    exposure_parser,
    gate_parser,
    image_output_parser,
    input_parser,
    interval_parser,
    trigger_parser,
    triggers,
    version_parser,
)


def determine_image_size(nexus_file: Path) -> tuple[int, int]:
    """Find the image size from metadata."""
    recommend = f"Please specify the detector dimensions in (x, y) with '--image-size'."
    try:
        with h5py.File(nexus_file) as f:
            # For the sake of some functions like zarr.create, ensure that the image
            # dimensions are definitely tuple[int, int], not tuple[np.int64,
            # np.int64] or anything else.
            y, x = f["entry/instrument/detector/module/data_size"][()].astype(int)
            return y, x
    except (FileNotFoundError, OSError):
        sys.exit(f"Cannot find NeXus file:\n\t{nexus_file}\n{recommend}")
    except KeyError:
        sys.exit(f"Input NeXus file does not contain image size metadata.\n{recommend}")


def exposure(
    start: int, end: int, exposure_time: pint.Quantity = None, num_images: int = None
) -> (pint.Quantity, int, int):
    """
    Find the exposure time or number of images.

    From a start time and an end time, either derive an exposure time from the
    number of images, or derive the number of images from the exposure time.

    Args:
        start:          Start time in clock cycles.
        end:            End time in clock cycles.
        exposure_time:  Exposure time in any unit of time (optional).
        num_images:     Number of images (optional).

    Returns:
        The exposure time in seconds, the exposure time in clock cycles, the number
        of images.
    """
    if exposure_time:
        exposure_time = exposure_time.to_base_units().to_compact()
        exposure_cycles = (exposure_time * clock_frequency).to_base_units().magnitude
        num_images = int((end - start) // exposure_cycles)
    else:
        # Because they are expected to be mutually exclusive, if there is no
        # exposure_time, there must be a num_images.
        exposure_cycles = int((end - start) / num_images)
        exposure_time = seconds(exposure_cycles)

    return exposure_time, exposure_cycles, num_images


def single_image_cli(args):
    """Utility for making a single image from event-mode data."""
    write_mode = "w" if args.force else "x"
    output_file = check_output_file(
        args.output_file, args.stem, "single_image", args.force
    )
    input_nexus = args.data_dir / f"{args.stem}.nxs"
    if input_nexus.exists():
        try:
            # Write output NeXus file if we have an input NeXus file.
            output_nexus = CopyTristanNexus.single_image_nexus(
                output_file, input_nexus, write_mode=write_mode
            )
        except FileExistsError:
            sys.exit(
                f"This output file already exists:\n\t"
                f"{output_file.with_suffix('.nxs')}\n"
                "Use '-f' to override, "
                "or specify a different output file path with '-o'."
            )
    else:
        output_nexus = None
        print(
            "Could not find a NeXus file containing experiment metadata.\n"
            "Resorting to writing raw image data without accompanying metadata."
        )

    image_size = args.image_size or determine_image_size(input_nexus)

    raw_files, _ = data_files(args.data_dir, args.stem)

    print("Finding detector shutter open and close times.")
    with latrd_data(raw_files, keys=cue_keys) as data, ProgressBar():
        start, end = find_start_end(data)

    print("Binning events into a single image.")
    with latrd_data(raw_files, keys=(event_location_key, event_time_key)) as data:
        data = valid_events(data, start, end)
        data[event_location_key] = pixel_index(data[event_location_key], image_size)
        image = da.bincount(data[event_location_key], minlength=mul(*image_size))
        image = image.astype(np.uint32).reshape(1, *image_size)

        with ProgressBar(), h5py.File(output_file, write_mode) as f:
            data_set = f.require_dataset(
                "data",
                shape=image.shape,
                dtype=image.dtype,
                chunks=image.chunksize,
                **Bitshuffle(),
            )
            image.store(data_set)

    print(f"Image written to\n\t{output_nexus or output_file}")


def multiple_images_cli(args):
    """
    Utility for making multiple images from event-mode data.

    The time between the start and end of the data collection is subdivided into a
    number of exposures of equal duration, providing a chronological stack of images.
    """
    write_mode = "w" if args.force else "x"
    output_file = check_output_file(args.output_file, args.stem, "images", args.force)

    input_nexus = args.data_dir / f"{args.stem}.nxs"
    if not input_nexus.exists():
        print(
            "Could not find a NeXus file containing experiment metadata.\n"
            "Resorting to writing raw image data without accompanying metadata."
        )

    image_size = args.image_size or determine_image_size(input_nexus)

    raw_files, _ = data_files(args.data_dir, args.stem)

    with latrd_data(raw_files, keys=cue_keys) as data:
        print("Finding detector shutter open and close times.")
        with ProgressBar():
            start, end = map(int, find_start_end(data))
        exposure_time, exposure_cycles, num_images = exposure(
            start, end, args.exposure_time, args.num_images
        )

        if args.align_trigger:
            trigger_type = triggers[args.align_trigger]
            print(
                f"Image start and end times will be chosen such that the first "
                f"'{cues[trigger_type]}' after the detector shutter open signal is "
                f"aligned with an image boundary."
            )
            # Note we are assuming that the first trigger time is after shutter open.
            trigger_time = first_cue_time(data, trigger_type, after=start)
            if trigger_time is None:
                sys.exit(
                    f"Could not find a '{cues[trigger_type]}' signal after the "
                    f"detector shutter open signal."
                )
            trigger_time = int(trigger_time.compute())

            if args.exposure_time:
                # Adjust the start time to align a bin edge with the trigger time.
                n_bins_before = (trigger_time - start) // exposure_cycles
                start = trigger_time - n_bins_before * exposure_cycles
                num_images = (end - start) // exposure_cycles
            # It is assumed that start ≤ trigger_time ≤ end.
            else:
                start, exposure_cycles = align_bins(
                    start, trigger_time, end, num_images
                )

        end = start + num_images * exposure_cycles
        bins = np.linspace(start, end, num_images + 1, dtype=np.uint64)

    if input_nexus.exists():
        try:
            # Write output NeXus file if we have an input NeXus file.
            output_nexus = CopyTristanNexus.multiple_images_nexus(
                output_file,
                input_nexus,
                nbins=num_images,
                write_mode=write_mode,
            )
        except FileExistsError:
            sys.exit(
                f"This output file already exists:\n\t"
                f"{output_file.with_suffix('.nxs')}\n"
                "Use '-f' to override, "
                "or specify a different output file path with '-o'."
            )
    else:
        output_nexus = None

    print(
        f"Binning events into {num_images} images with an exposure time of "
        f"{exposure_time:.3g~#P}."
    )

    # Make a cache for the images.
    images = create_cache(output_file, num_images, image_size)

    with latrd_data(raw_files, keys=(event_location_key, event_time_key)) as data:
        data = events_to_images(data, bins, image_size, images)

        print("Computing the binned images.")
        # Use multi-threading, rather than multi-processing.
        with Client(processes=False):
            compute_with_progress(data)

    print("Transferring the images to the output file.")
    with h5py.File(output_file, write_mode) as f:
        zarr.copy_all(zarr.open(images.store), f, **Bitshuffle())

    # Delete the Zarr store.
    images.store.clear()

    print(f"Images written to\n\t{output_nexus or output_file}")


def pump_probe_cli(args):
    """
    Utility for making multiple images from a pump-probe data collection.

    The time between one pump trigger signal and the next is subdivided into a number
    of exposures of equal duration.  Data from all such pump-to-pump intervals is
    aggregated, providing a single stack of images that captures the evolution of the
    response of the measurement to a pump signal.
    """
    write_mode = "w" if args.force else "x"
    output_file = check_output_file(args.output_file, args.stem, "images", args.force)

    input_nexus = args.data_dir / f"{args.stem}.nxs"
    if not input_nexus.exists():
        print(
            "Could not find a NeXus file containing experiment metadata.\n"
            "Resorting to writing raw image data without accompanying metadata."
        )

    image_size = args.image_size or determine_image_size(input_nexus)

    raw_files, _ = data_files(args.data_dir, args.stem)

    trigger_type = triggers.get(args.trigger_type)

    with latrd_data(raw_files, keys=cue_keys) as cues_data:
        print("Finding trigger signal times.")
        trigger_times = cue_times(cues_data, trigger_type)
        with ProgressBar():
            # Assumes the trigger times can be held in memory.
            trigger_times = trigger_times.astype(int).compute()

    if not trigger_times.size:
        sys.exit(f"Could not find a '{cues[trigger_type]}' signal.")
    elif not trigger_times.size > 1:
        sys.exit(f"Only one '{cues[trigger_type]}' signal found.  Two or more needed.")

    end = da.diff(trigger_times).min()
    exposure_time, num_images = args.exposure_time, args.num_images
    exposure_time, _, num_images = exposure(0, end, exposure_time, num_images)
    bins = np.linspace(0, end, num_images + 1, dtype=np.uint64)

    if input_nexus.exists():
        try:
            # Write output NeXus file if we have an input NeXus file.
            output_nexus = CopyTristanNexus.single_image_nexus(
                output_file,
                input_nexus,
                write_mode=write_mode,
                pump_probe_bins=num_images,
            )
        except FileExistsError:
            sys.exit(
                f"This output file already exists:\n\t"
                f"{output_file.with_suffix('.nxs')}\n"
                "Use '-f' to override, "
                "or specify a different output file path with '-o'."
            )
    else:
        output_nexus = None

    print(
        f"Binning events into {num_images} images with an exposure time of "
        f"{exposure_time:.3g~#P} according to the time elapsed since the most "
        f"recent '{cues[trigger_type]}' signal."
    )

    # Make a cache for the images.
    images = create_cache(output_file, num_images, image_size)

    # Get the events data.
    keys = (event_location_key, event_time_key)
    with latrd_data(raw_files, keys=keys) as events_data:
        # Measure the event time as time elapsed since the most recent trigger signal.
        events_data = events_data.astype({event_time_key: np.int64})
        event_times = events_data[event_time_key].values
        trigger_index = da.digitize(event_times, trigger_times) - 1
        events_data[event_time_key] -= da.take(trigger_times, trigger_index)

        # Bin the events into images.
        events_data = events_to_images(events_data, bins, image_size, images)

        print("Computing the binned images.")
        # Use multi-threading, rather than multi-processing.
        with Client(processes=False):
            compute_with_progress(events_data)

    print("Transferring the images to the output file.")
    with h5py.File(output_file, write_mode) as f:
        zarr.copy_all(zarr.open(images.store), f, **Bitshuffle())

    # Delete the Zarr store.
    images.store.clear()

    print(f"Images written to\n\t{output_nexus or output_file}")


def multiple_sequences_cli(args):
    """
    Utility for making multiple image sequences from a pump-probe data collection.

    The time between one pump trigger signal and the next is subdivided into a number
    of intervals of equal duration, quantising the time elapsed since the most recent
    trigger pulse.  Events are labelled according to the interval into which they fall
    and, for each interval in turn, all the events so labelled are binned into a
    sequence of images, providing a stack of image sequences that captures the
    evolution of the response of the measurement to a pump signal.
    """
    write_mode = "w" if args.force else "x"

    input_nexus = args.data_dir / f"{args.stem}.nxs"
    if not input_nexus.exists():
        print(
            "Could not find a NeXus file containing experiment metadata.\n"
            "Resorting to writing raw image data without accompanying metadata."
        )

    image_size = args.image_size or determine_image_size(input_nexus)

    raw_files, _ = data_files(args.data_dir, args.stem)

    trigger_type = triggers.get(args.trigger_type)

    print("Finding trigger signal times.")

    with latrd_data(raw_files, keys=cue_keys) as cues_data:

        trigger_times = cue_times(cues_data, trigger_type)
        with ProgressBar():
            trigger_times = trigger_times.astype(int).compute()

        if not trigger_times.size:
            sys.exit(f"Could not find a '{cues[trigger_type]}' signal.")
        elif not trigger_times.size > 1:
            sys.exit(
                f"Only one '{cues[trigger_type]}' signal found.  Two or more needed."
            )

        print("Finding detector shutter open and close times.")
        with ProgressBar():
            start, end = find_start_end(cues_data)

    intervals_end = da.diff(trigger_times).min()
    interval_time, _, num_intervals = exposure(
        0, intervals_end, args.interval, args.num_sequences
    )
    # Find the bins denoting to which image sequence each event belongs.
    interval_bins = np.linspace(0, intervals_end, num_intervals + 1, dtype=np.uint64)

    output_files, out_file_pattern = check_multiple_output_files(
        num_intervals, args.output_file, args.stem, "images", args.force
    )

    exposure_time, exposure_cycles, num_images = exposure(
        start, end, args.exposure_time, args.num_images
    )
    # Find the bins denoting images within a sequence.
    bins = np.linspace(start, end, num_images + 1, dtype=np.uint64)

    print(
        f"Using '{cues[trigger_type]}' as the pump signal,\n"
        f"binning events into {num_intervals} sequences, corresponding to "
        f"successive pump-probe delay intervals of {interval_time:.3g~#P}.\n"
        f"Each sequence consists of {num_images} images with an effective exposure "
        f"time of {exposure_time / num_intervals:.3g~#P}."
    )

    out_file_stem = out_file_pattern.stem

    n_dig = len(str(num_intervals))
    out_file_pattern = out_file_pattern.parent / f"{out_file_stem}_{'#' * n_dig}.h5"

    if input_nexus.exists():
        # Write output NeXus files if we have an input NeXus file.
        output_nexus_pattern = out_file_pattern.with_suffix(".nxs")
        for output_file in output_files:
            try:
                CopyTristanNexus.multiple_images_nexus(
                    output_file,
                    input_nexus,
                    nbins=num_images,
                    write_mode=write_mode,
                )
            except FileExistsError:
                sys.exit(
                    f"One or more output files already exist, "
                    f"matching the pattern:\n\t"
                    f"{output_nexus_pattern}\n"
                    "Use '-f' to override, "
                    "or specify a different output file path with '-o'."
                )
    else:
        output_nexus_pattern = None

    # Make a cache for the images.
    images = create_cache(out_file_pattern, num_intervals * num_images, image_size)

    # Get the events data.
    events_keys = (event_location_key, event_time_key)
    with latrd_data(raw_files, keys=events_keys) as events_data:
        events_data = valid_events(events_data, start, end)

        # Find the time elapsed since the most recent trigger signal.
        event_time = events_data[event_time_key].astype(np.int64).values
        trigger_index = da.digitize(event_time, trigger_times) - 1
        pump_probe_time = event_time - da.take(trigger_times, trigger_index)
        # Enumerate the sequence to which each event belongs.
        sequence = da.digitize(pump_probe_time, interval_bins) - 1
        # Eliminate invalid sequence numbers (negative, or ≥ num_intervals).
        valid = (0 <= sequence) & (sequence < num_intervals)
        valid = dd.from_dask_array(valid, index=events_data.index)

        # Convert the event IDs to a form that is suitable for a NumPy bincount.
        events_data[event_location_key] = pixel_index(
            events_data[event_location_key], image_size
        )

        columns = event_location_key, "time_bin"
        dtypes = events_data.dtypes
        dtypes["time_bin"] = dtypes.pop(event_time_key)
        meta = pd.DataFrame(columns=columns).astype(dtype=dtypes)
        # Enumerate the image in the stack to which each event belongs.
        events_data = events_data.map_partitions(find_time_bins, bins=bins, meta=meta)
        events_data["time_bin"] += sequence * num_images
        events_data = events_data[valid]

        # Bin to images, partition by partition.
        events_data = dd.map_partitions(
            make_images,
            events_data,
            image_size,
            images,
            meta=meta,
            enforce_metadata=False,
        )
        print("Computing the binned images.")
        # Use multi-threading, rather than multi-processing.
        with Client(processes=False):
            compute_with_progress(events_data)

    print("Transferring the images to the output files.")
    store = images.store
    images = da.from_zarr(images)
    stack_shape = num_intervals, num_images, *image_size
    # Silence a large chunks warning, since we immediately rechunk to one-image chunks.
    with dask.config.set(**{"array.slicing.split_large_chunks": False}):
        images = images.reshape(stack_shape).rechunk((1, 1, *image_size))
    images = list(images)

    # Multi-threaded copy from Zarr to HDF5.
    with ExitStack() as stack:
        files = (stack.enter_context(h5py.File(f, write_mode)) for f in output_files)
        dsets = [
            f.require_dataset(
                "data",
                shape=images[0].shape,
                dtype=images[0].dtype,
                chunks=images[0].chunksize,
                **Bitshuffle(),
            )
            for f in files
        ]
        with ProgressBar():
            da.store(images, dsets)

    # Delete the Zarr store.
    store.clear()

    print(f"Images written to\n\t{output_nexus_pattern or out_file_pattern}")


def gated_images_cli(args):
    """Utility to bin events into a sequence of images according to a gating signal."""
    write_mode = "w" if args.force else "x"
    output_file = check_output_file(args.output_file, args.stem, "images", args.force)

    input_nexus = args.data_dir / f"{args.stem}.nxs"
    if not input_nexus.exists():
        print(
            "Could not find a NeXus file containing experiment metadata.\n"
            "Resorting to writing raw image data without accompanying metadata."
        )

    image_size = args.image_size or determine_image_size(input_nexus)

    raw_files, _ = data_files(args.data_dir, args.stem)

    # If gate_close isn't specified, default to the complementary signal to gate_open.
    gate_open = triggers.get(args.gate_open)
    gate_close = triggers.get(args.gate_close) or gate_open ^ (1 << 5)

    with latrd_data(raw_files, keys=cue_keys) as cues_data:
        print("Finding detector shutter open and close times.")
        with ProgressBar():
            start, end = find_start_end(cues_data)

        print("Finding gate signal times.")
        # Here we assume no synchronization issues:
        # falling edges always recorded after rising edges.
        open_times = cue_times(cues_data, gate_open, after=start)
        close_times = cue_times(cues_data, gate_close, before=end)
        with ProgressBar():
            open_times, close_times = dask.compute(open_times, close_times)

    if not open_times.size:
        sys.exit(f"Could not find a '{cues[gate_open]}' signal.")
    if not close_times.size:
        sys.exit(f"Could not find a '{cues[gate_close]}' signal.")

    open_times = np.sort(open_times)
    close_times = np.sort(close_times)

    if not open_times.size == close_times.size:
        # If size difference is just one, look for missing one right before/after
        # shutters and use shutter open/close timestamp as first/last gate
        if abs(open_times.size - close_times.size) > 1:
            sys.exit(
                "Found a non-matching number of gate open and close signals:\n\t"
                f"Number of '{cues[gate_open]}' signals: {open_times.size}\n\t"
                f"Number of '{cues[gate_close]}' signals: {close_times.size}\n"
                f"Note that signals before the shutter open time are ignored."
            )
        else:
            if open_times[-1] > close_times[-1]:
                print(
                    "WARNING! \n\t"
                    f"Missing last '{cues[gate_close]}' signal.\n\t"
                    f"Shutter close timestamp will be used instead for last image."
                )
                # Append shutter close to close_times
                close_times = np.append(close_times, end)
            elif open_times[0] > close_times[0]:
                print(
                    "WARNING! \n\t"
                    f"Missing first '{cues[gate_open]}' signal.\n\t"
                    f"Shutter open timestamp will be used instead for first image."
                )
                # Insert shutter open to open times
                open_times = np.insert(open_times, 0, start)
            else:
                sys.exit(
                    "Found a non-matching number of gate open and close signals:\n\t"
                    f"Number of '{cues[gate_open]}' signals: {open_times.size}\n\t"
                    f"Number of '{cues[gate_close]}' signals: {close_times.size}\n"
                )

    num_images = open_times.size
    bins = np.linspace(0, num_images, num_images + 1, dtype=np.uint64)

    if input_nexus.exists():
        try:
            # Write output NeXus file if we have an input NeXus file.
            output_nexus = CopyTristanNexus.serial_images_nexus(
                output_file,
                input_nexus,
                nbins=num_images,
                write_mode=write_mode,
            )
        except FileExistsError:
            sys.exit(
                f"This output file already exists:\n\t"
                f"{output_file.with_suffix('.nxs')}\n"
                "Use '-f' to override, "
                "or specify a different output file path with '-o'."
            )
    else:
        output_nexus = None

    print(f"Binning events into {num_images} images.")

    # Make a cache for the images.
    images = create_cache(output_file, num_images, image_size)

    with latrd_data(raw_files, keys=(event_location_key, event_time_key)) as data:
        # Consider only those events that occur between the start and end times.
        data = valid_events(data, start, end)

        # Gate the events.
        event_times = data[event_time_key].astype(np.int64).values
        open_index = da.digitize(event_times, open_times) - 1
        close_index = da.digitize(event_times, close_times)
        # Look for events that happen after gate open and before gate close
        # Eliminate invalid events by looking at the open and close index
        valid = open_index == close_index
        valid = dd.from_dask_array(valid, index=data.index)

        # Convert the event IDs to a form that is suitable for a NumPy bincount.
        data[event_location_key] = pixel_index(data[event_location_key], image_size)

        columns = event_location_key, "time_bin"
        dtypes = data.dtypes
        dtypes["time_bin"] = dtypes.pop(event_time_key)

        meta = pd.DataFrame(columns=columns).astype(dtype=dtypes)
        # Enumerate the image in the stack to which each event belongs
        data = data.map_partitions(find_time_bins, bins=bins, meta=meta)
        data["time_bin"] = open_index
        data = data[valid]

        # Bin to images, partition by partition.
        data = dd.map_partitions(
            make_images, data, image_size, images, meta=meta, enforce_metadata=False
        )

        print("Computing the binned images.")
        # Use multi-threading, rather than multi-processing.
        with Client(processes=False):
            compute_with_progress(data)

    print("Transferring the images to the output file.")
    with h5py.File(output_file, write_mode) as f:
        zarr.copy_all(zarr.open(images.store), f, **Bitshuffle())

    # Delete the Zarr store.
    images.store.clear()

    print(f"Images written to\n\t{output_nexus or output_file}")


parser = argparse.ArgumentParser(description=__doc__, parents=[version_parser])
subparsers = parser.add_subparsers(
    help="Choose the manner in which to create images.",
    required=True,
    dest="sub-command",
)

parser_single = subparsers.add_parser(
    "single",
    aliases=["1"],
    description=(
        "Aggregate all the events from a LATRD Tristan data collection "
        "into a single image."
    ),
    parents=[version_parser, input_parser, image_output_parser],
)
parser_single.set_defaults(func=single_image_cli)

parser_multiple = subparsers.add_parser(
    "multiple",
    aliases=["multi"],
    description=(
        "Bin the events from a LATRD Tristan data collection into multiple images."
    ),
    parents=[version_parser, input_parser, image_output_parser, exposure_parser],
)
parser_multiple.add_argument(
    "-a",
    "--align-trigger",
    help="Align the start and end time of images such that the first trigger signal of "
    "the chosen type is matched up with an image start time.  Useful for examining "
    "effects in the data before and after a single trigger pulse.",
    choices=triggers.keys(),
)
parser_multiple.set_defaults(func=multiple_images_cli)

parser_pump_probe = subparsers.add_parser(
    "pump-probe",
    aliases=["pp"],
    description="Bin events into images representing different pump-probe delays.\n\n"
    "With LATRD data from a pump-probe experiment, where the pump signal has a fairly "
    "constant repeat rate, bin events into a stack of images spanning the range of "
    "pump-probe delay times, from shortest to longest.",
    parents=[
        version_parser,
        input_parser,
        image_output_parser,
        trigger_parser,
        exposure_parser,
    ],
)
parser_pump_probe.set_defaults(func=pump_probe_cli)

parser_multiple_sequences = subparsers.add_parser(
    "sequences",
    aliases=["sweeps"],
    description="Bin events into several sequences of images, each corresponding to "
    "a different pump-probe delay time interval.\n\n"
    "With LATRD data from a pump-probe experiment, where the pump signal has a fairly "
    "constant repeat rate, the recorded events are separated into groups corresponding "
    "to the time elapsed since the most recent pump trigger signal.  Each group is "
    "binned into a sequence of chronological images.  Each sequence is saved to a "
    "separate output file, numbered from the shortest pump-probe delay to the longest.",
    parents=[
        version_parser,
        input_parser,
        image_output_parser,
        trigger_parser,
        exposure_parser,
        interval_parser,
    ],
)
parser_multiple_sequences.set_defaults(func=multiple_sequences_cli)

parser_serial = subparsers.add_parser(
    "serial",
    description="Bin events into images, gated with trigger signals.\n\n"
    "Events will be binned into as many images as there are gate signals, one image "
    "per gate.  Each 'gate-open' signal is taken as the start of an exposure and the "
    "next 'gate-close' signal is taken as the end of the exposure.",
    parents=[version_parser, input_parser, image_output_parser, gate_parser],
)
parser_serial.set_defaults(func=gated_images_cli)


def main(args=None):
    """Perform the image binning with a user-specified sub-command."""
    args = parser.parse_args(args)
    args.func(args)
